// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
//       Free Software Foundation, Inc.
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>
#include <sstream>

#include <math.h>

#include <matvec/macros.h>
#include <randlib_c/randlib.h>

#include  "scatter.h"


// use matvec macros, but simplify them

#define VEC(vec) MATVEC_VEC_AS_STRUCT(vec,x,y,z)

#define VCROSS(p, v1, v2)				\
    MATVEC_VEC_CROSSPROD( VEC(p), VEC(v1), VEC(v2) )

#define VDOT(v1, v2)				\
    MATVEC_VEC_DOTPROD( VEC(v1), VEC(v2))

#define VLEN(v) sqrt(VDOT((v),(v)))

#define VADD( v, v1, v2 )			\
    MATVEC_VEC_ADD( VEC(v), VEC(v1), VEC(v2) )

#define VADDTO( v1, v2 )			\
    MATVEC_VEC_ADD( VEC(v1), VEC(v1), VEC(v2) )

#define VSET( v1, v2 )				\
    MATVEC_VEC_ASSIGN( VEC(v1), VEC(v2 ) )

#define VSCALE( r, v, c )			\
    MATVEC_VEC_SCALE( VEC(r), VEC(v), c )

#define VNORM(v)				\
    MATVEC_VEC_NORM(VEC(v))

using namespace Scatter;

unsigned int
scatter(
	const ZoneMap::Zone* zone,
	int  iter_limit,
	double theta_def, double phi_def,
	double min_cum_prob, double max_cum_prob,
	const unsigned int id,
	const double energy,
	const DVector3& surface_norm,
	DVector3& direction,
	double& theta_o,
	double& phi_o
	)
{

    double sin_alpha  = -1.0 * VDOT( surface_norm, direction );
    double cos_alpha  = sqrt( 1.0 - sin_alpha * sin_alpha );

    // default scatter angles; convert from arcseconds to radians
    double theta      = theta_def;
    double phi        = phi_def;

    double delta_prob = max_cum_prob - min_cum_prob;

    unsigned int status = OK;


#ifdef DEBUG
    double alpha	    = asin( sin_alpha );
#endif // DEBUG

    // temporary direction vector to handle possible iterations
    DVector3 tdir;

    do
    {
	// so far, nothing bad has happened
	status = OK;

	tdir = direction;

	if ( zone->in_plane ) {

	    bool is_extrapolated =
		zone->db_in_plane->lookUpScatterAngle( energy, sin_alpha,
						       min_cum_prob + ranf( ) * delta_prob,
						       theta );
	    if ( is_extrapolated )
		status |= IP_EXTRAPOLATED;

	    theta *= zone->scale_in_plane;
	}

	if ( zone->out_of_plane ) {

	    bool is_extrapolated =
		zone->db_out_of_plane->lookUpScatterAngle( energy, sin_alpha,
							   min_cum_prob + ranf( ) * delta_prob,
							   phi );
	    if ( is_extrapolated )
		status |= OP_EXTRAPOLATED;

	    if ( zone->scale_out_of_plane_by_sin_alpha )
		phi *= sin_alpha;

	    if (  zone->make_out_of_plane_symmetric
		  && ranf() > 0.5 )
			phi *= -1;

	    phi *= zone->scale_out_of_plane;
	}

	theta_o  = theta;
	phi_o = phi;

	double cos_theta = cos( theta );
	double sin_theta = sin( theta );
	double cos_phi   = cos( phi   );
	double sin_phi   = sin( phi   );

	// cos(alpha+theta)
	double cos_at = cos_alpha * cos_theta - sin_alpha * sin_theta;
	// sin(alpha+theta)
	double sin_at = sin_alpha * cos_theta + cos_alpha * sin_theta;

	double A = cos_at / cos_alpha;
	double B = A * sin_alpha * cos_phi - sin_at;

	DVector3 r_cross_n;
	VCROSS(r_cross_n, direction, surface_norm );

	DVector3 v;
	VSCALE(v, surface_norm, sin_alpha);
	VADDTO(v, direction);
	VSCALE(v, v, 1/cos_alpha );

#ifdef DEBUG
	DVector3 u;

	VCROSS( u, direction, surface_norm );
	VNORM(u);

	//DVector3 v;
	//VCROSS( v, surface_norm, u );

#endif // DEBUG


	/* Construct new reflected ray. note that it's a unit vector */

	tdir.x = B * surface_norm.x + A * (cos_phi * direction.x + sin_phi * r_cross_n.x);
	tdir.y = B * surface_norm.y + A * (cos_phi * direction.y + sin_phi * r_cross_n.y);
	tdir.z = B * surface_norm.z + A * (cos_phi * direction.z + sin_phi * r_cross_n.z);

	// not quite normalized due to numerics. fix that.
	VNORM(tdir);

	// has the ray been scattered into the reflecting surface?
	if ( VDOT( tdir, surface_norm ) > 0 )
	    status |= SCATTERED_INTO_SURFACE;

	// has the ray been backscattered? check that the v components
	// have the same sign (i.e. they're on the same side of the
	// surface normal in the specular plane.  if they have the same
	// sign their product will be >= 0;
	if ( VDOT( tdir, v) * VDOT( direction, v ) < 0 )
	    status |= BACKSCATTERED;

#ifdef DEBUG

	if ( dbflag( "ray" ) )
	    tf.message( "ray:\t(%.4f %.4f %.4f) -> (%.4f %.4f %.4f)\n",
			direction.x, direction.y, direction.z, tdir.x, tdir.y, tdir.z );

	if ( dbflag( "norm" ) )
	    tf.message( "norm:\t(%.4f %.4f %.4f)\n", surface_norm.x, surface_norm.y, surface_norm.z );

	if ( dbflag( "theta" ) )
	{
	    double theta_out = asin( - VDOT(tdir, surface_norm ) ) - alpha;
	    tf.message( "theta_in:\t%u\t%.14g\n", id, RAD2ARCSEC(theta) );
	    tf.message( "theta_out:\t%u\t%.14g\n", id, RAD2ARCSEC(theta_out) );
	}

	if ( dbflag( "phi" ) )
	{
	    DVector3 _v;
	    DVector3 _u;

	    VSCALE( _v, v, VDOT(tdir, v) );

	    VSCALE( _u, u, VDOT(tdir, u) );

	    DVector3 l;

	    VADD( l, _v, _u );

	    VNORM(l);

	    tf.message( "phi_in:\t\t%u\t%.14g\n", id, RAD2ARCSEC(phi) );
	    tf.message( "phi_out:\t%u\t%.14g\n", id, RAD2ARCSEC(acos( VDOT(l, v ) ) * ( VDOT(l,u) < 0 ? -1 : 1 )) );
	}

#endif // DEBUG


    } while( --iter_limit && status & SCATTERED_INTO_SURFACE );

    direction = tdir;

    return status;
}

